/******************************************************************************
* bandwidth_monitor.c - $Revision: 1.4 $ - $Date: 2004/11/14 18:38:31 $
*
*   Copyright (C) 2004 Craig B. Agricola (craig@theagricolas.org)
*
* This program started as a shell script that I ran on my home-brew firewall
* wireless router (Linux, of course, on a 486/66) to keep track of my incoming
* and outgoing bandwidth.  This let me know that I was getting the bandwidth
* that I paid for, and it also helped me keep an eye on the throttling of
* bulk traffic so that I didn't choke out my interactive sessions.  Of course,
* that was before I had QOS.
*
* Anyway, before long, I re-coded it as a Perl script to allow a bit more
* flexibility, and used it like that for quite a long time.  Then I got a
* Linksys WRT54GS to replace my firewall, and realized that I really missed
* having my bandwidth stats streaming to a terminal window.  Of course, putting
* Perl on the WRT54GS is just overkill, so I re-coded it again in C.  Now it
* compiles and runs very nicely on any Linux box that has a 2.4 or beyond
* kernel, as well as embedded Linux systems.
*
* The program is probably tied to Linux, since it grabs the byte counts for the
* requested interface right from /proc/net/dev.  It could pretty easily be
* modified to grab the same information from the output of /sbin/ifconfig, but
* that would waste resources to spawn a new process to parse the /proc/net/dev
* file and regurgitate it to us.  Also, I haven't bothered to look at how
* uniform the output of the byte counts from /sbin/ifconfig is on other
* operating systems.  As a side note, the C version also uses nanosleep() and
* gettimeofday() to get more accurate polling intervals and track the overall
* time (since sleep(), usleep(), nanosleep(), et. al. only guaruntee to wait
* at least as long as you request, but do not guaruntee that it won't be
* possibly much longer...)  So, if you were interested in porting it, you'd
* have to make sure that nanosleep() and gettimeofday() were also available,
* or lose the resolution of the time used in bandwidth calculations.
* 
* And finally, if you actually pay attention, you will notice that the
* bandwidth numbers reported are greater than the bandwidth you will perceive
* on a file transfer.  This is not a bug, as you may presume it to be.  This
* is because the bandwidth being measured includes all of the protocol and TCP
* overhead.  If you just time a transfer of a file and then divide the size
* of the file by the time, you get the usable bandwidth, but don't see any of
* the overhead bandwidth.  Of course, the bandwidth that your ISP advertises
* includes the overhead.  This is why if you have a 3Mbps downlink from your
* cable company, you will never be able to download a 3MB file in 8 seconds
* (remember the factor of 8; they advertise *bits* per second, not *bytes*
* per second!)
*
* BUILDING
*
* This is how I build it with GCC under X86 Linux.
*
* gcc -O3 -o bandwidth_monitor bandwidth_monitor.c
*
* This is how I build it with GCC under X86 Linux for a uclibc/Linux MIPS
* embedded machine (like the Linksys WRT54G/S).  Obviously this requires a
* cross-compiler.
*
* mipsel-linux-gcc -O2 -static -o bandwidth_monitor bandwidth_monitor.c
* mipsel-linux-strip bandwidth_monitor
*
* If you don't want to bother with building a cross-compiler, and if you
* feel like trusting me, you can ask my for my current build at
* craig@theagricolas.org
*
* I'd love feedback on how it works with other embedded targets and operating
* systems.
*
* USAGE
*
* bandwidth_monitor <device> <polling_interval> <number_of_polls>
*
* <device> is the network device that you want to monitor the bandwidth of.
*          The default is eth0, but on my WRT54GS with Sveasoft's Satori-4.0
*          v2.07.1.7sv installed, the internet side of the router seems to be
*          vlan1.
* <polling_interval> is the number of seconds between polls of the byte counts.
*                    This is a floating point number, so you can poll much more
*                    quickly than once a second if you want.  The default is
*                    one second, though I tend to find that 3 seconds is good
*                    for smoothing out burstiness in my internet usage.
* <number_of_polls> is how many times it will poll before exiting.  When it
*                   exits, it will calculate the average bandwidths over the
*                   entire time that it has run.  The default is -1, which means
*                   that it will run until it is killed.  A SIGINT (control-C)
*                   will kill it, but the overall average bandwidth will still
*                   be output.
*
* LICENSE
*
* bandwidth_monitor is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation; either version 2, or (at your option) any later version.
*
* bandwidth_monitor is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
* details.
*
* You should be able to download a copy of the GNU General Public License from
* http://www.gnu.org/licenses/gpl.txt
* If not, write to the Free Software Foundation,
* 59 Temple Place, Suite 330, Boston, MA 02111 USA.
*
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <limits.h>

static int quit = 0;
void sighdl(int sig) { quit = 1; }

struct byte_counts {
  struct timeval time_stamp;
  unsigned int   rx_bytes;
  unsigned int   tx_bytes;
};

void copy_byte_counts(struct byte_counts *src, struct byte_counts *dst) {
  memcpy(dst, src, sizeof(struct byte_counts));
}

int get_byte_counts(char *dev, struct byte_counts *counts) {
  char buf[150], *b;
  char *line_dev;
  int c, found;
  FILE *f;
  f = fopen("/proc/net/dev", "r");
  if (!f) return(0);
  for(c=0; c<2; c++) fgets(buf, 150, f);
  found = 0;
  while(!found && fgets(buf, 150, f)) {
    for(b=buf; *b && *b==' '; b++);
    line_dev = b;
    for(; *b && *b!=':'; b++);
    *b++=0;
    if (!strcmp(dev, line_dev)) found=1;
  }
  if (found && counts) {
    gettimeofday(&counts->time_stamp, 0);
    for(; *b && *b==' '; b++);
    sscanf(b, "%u", &counts->rx_bytes);
    for(c=0; c<8; c++) {
      for(; *b && *b!=' '; b++);
      for(; *b && *b==' '; b++);
    }
    sscanf(b, "%u", &counts->tx_bytes);
  }
  fclose(f);
  return(found);
}

char *bandwidth_units[] = { " bps", "Kbps", "Mbps", "Gbps" };
void calculate_bandwidth(double delay, unsigned int start, unsigned int end,
                         double *bw, char **bwu)
{
  int bwui;
  if (end >= start)
    *bw = (end - start) * 8 / delay;
  else
    *bw = ((UINT_MAX - start) + end) * 8 / delay;
  for(bwui=0; (bwui<4) && (*bw > 1024); bwui++) *bw /= 1024;
  *bwu = bandwidth_units[bwui];
}

double calculate_timespan(struct timeval *start, struct timeval *end) {
  double timespan;
  timespan = end->tv_sec - start->tv_sec;
  if (end->tv_usec < start->tv_usec) {
    timespan -= 1;
    timespan += (((double)(1000000 - start->tv_usec + end->tv_usec)) / 1000000);
  } else
    timespan += (((double)(end->tv_usec - start->tv_usec)) / 1000000);
  return(timespan);
}

void log_bandwidth(struct byte_counts *start, struct byte_counts *end) {
  double span;
  double rxbw, txbw;
  char *rxu, *txu;

  span = calculate_timespan(&start->time_stamp, &end->time_stamp);
  calculate_bandwidth(span, start->rx_bytes, end->rx_bytes, &rxbw, &rxu);
  calculate_bandwidth(span, start->tx_bytes, end->tx_bytes, &txbw, &txu);
  printf("Receive: %9.3f %s     Transmit: %9.3f %s\n",
         rxbw, rxu, txbw, txu);
}

void fractional_sleep(double seconds) {
  struct timespec ts;
  ts.tv_sec = (int)seconds;
  ts.tv_nsec = (seconds - (double)((int)seconds)) * 1000000000;
  nanosleep(&ts, 0);
}

int main(int argc, char *argv[]) {
  struct byte_counts start;
  struct byte_counts last;
  struct byte_counts now;
  char *dev;
  double delay;
  int end_count, count;

  if (argc<2) dev="eth1";
    else      dev=argv[1];
  if (argc<3) delay=1.0;
    else      delay=strtod(argv[2], 0);
  if (argc<4) end_count=-1;
    else      end_count=strtol(argv[3], 0, 0);

  if (!get_byte_counts(dev, &start)) return(-1);
  copy_byte_counts(&start, &last);
  copy_byte_counts(&start, &now);
  signal(SIGINT, sighdl);
  for(count=0; !quit && ((end_count<0)||(count<end_count)); count++) {
    fractional_sleep(delay);
    if (get_byte_counts(dev, &now)) {
      log_bandwidth(&last, &now);
      copy_byte_counts(&now, &last);
    }
  }
  printf("\nOverall Average:\n");
  log_bandwidth(&start, &last);
  return(0);
}
